import warnings

import numpy as np
import os
import shutil
import sys

import hcai_datasets.hcai_nova_dynamic.utils.nova_types as nt
import hcai_datasets.hcai_nova_dynamic.utils.nova_data_utils as ndu
import hcai_datasets.hcai_nova_dynamic.utils.nova_anno_utils as nau
from hcai_dataset_utils.dataset_iterable import DatasetIterable

from hcai_datasets.hcai_nova_dynamic.utils.nova_data_utils import (
    AudioData,
    VideoData,
    StreamData,
)
from hcai_datasets.hcai_nova_dynamic.utils.nova_anno_utils import (
    DiscreteAnnotation,
    ContinuousAnnotation,
    FreeAnnotation,
)

from hcai_datasets.hcai_nova_dynamic.nova_db_handler import NovaDBHandler
from hcai_datasets.hcai_nova_dynamic.utils.nova_utils import *


class HcaiNovaDynamicIterable(DatasetIterable):
    def __init__(
            self,
            *,
            db_config_path=None,
            db_config_dict=None,
            dataset=None,
            nova_data_dir=None,
            sessions=None,
            annotator=None,
            schemes=None,
            roles=None,
            data_streams=None,
            start=None,
            end=None,
            left_context="0s",
            right_context="0s",
            frame_size=None,
            stride=None,
            flatten_samples=False,
            supervised_keys=None,
            add_rest_class=True,
            lazy_loading=False,
            **kwargs
    ):
        """
        Initialize the HcaiNovaDynamic dataset builder
        Args:
          nova_data_dir: the directory to look for data. same as the directory specified in the nova gui.
          frame_size: the framesize to look at. the matching annotation will be calculated as majority vote from all annotations that are overlapping with the timeframe.
          left_context: additional data to pass to the classifier on the left side of the frame.
          right_context: additional data to pass to the classifier on the left side of the frame.
          stride: how much a frame is moved to calculate the next sample. equals framesize by default.
          flatten_samples: if set to True samples with the same annotation scheme but from different roles will be treated as separate samples. only <scheme> is used for the keys.
          supervised_keys: if specified the dataset can be used with "as_supervised" set to True. Should be in the format <role>.<scheme>. if flatten_samples is true <role> will be ignored.
          add_rest_class: when set to True an additional restclass will be added to the end the label list
          db_config_path: path to a configfile whith the nova database config.
          db_config_dict: dictionary with the nova database config. can be used instead of db_config_path. if both are specified db_config_dict is used.
          dataset: the name of the dataset. must match the dataset name in the nova database.
          sessions: list of sessions that should be loaded. must match the session names in nova.
          annotator: the name of the annotator that labeld the session. must match annotator names in nova.
          schemes: list of the annotation schemes to fetch
          roles: list of roles for which the annotation should be loaded.
          data_streams: list datastreams for which the annotation should be loaded. must match stream names in nova.
          start: optional start time_ms. use if only a specific chunk of a session should be retreived.
          end: optional end time_ms. use if only a specifc chunk of a session should be retreived.
          **kwargs: arguments that will be passed through to the dataset builder
        """
        self.dataset = dataset
        self.nova_data_dir = nova_data_dir
        self.sessions = sessions
        self.roles = roles
        self.schemes = schemes
        self.data_streams = data_streams
        self.annotator = annotator
        self.left_context_ms = ndu.parse_time_string_to_ms(left_context)
        self.right_context_ms = ndu.parse_time_string_to_ms(right_context)
        self.frame_size_ms = ndu.parse_time_string_to_ms(frame_size) if frame_size else None
        self.stride_ms = (
            ndu.parse_time_string_to_ms(stride) if stride else self.frame_size_ms
        )
        self.start_ms = ndu.parse_time_string_to_ms(start)
        if not self.start_ms:
            self.start_ms = 0

        self.end_ms = ndu.parse_time_string_to_ms(end)
        if not self.end_ms:
            self.end_ms = float("inf")

        self.flatten_samples = flatten_samples
        self.add_rest_class = add_rest_class
        self.lazy_loading = lazy_loading
        self.nova_db_handler = NovaDBHandler(db_config_path, db_config_dict)

        # Retrieving meta information from the database
        mongo_schemes = self.nova_db_handler.get_schemes(
            dataset=dataset, schemes=schemes
        )
        mongo_data = self.nova_db_handler.get_data_streams(
            dataset=dataset, data_streams=data_streams
        )
        self.annos, self.anno_schemes = self._populate_label_info_from_mongo_doc(mongo_schemes)
        self.data_info, self.data_schemes = self._populate_data_info_from_mongo_doc(mongo_data)

        # setting supervised keys
        if supervised_keys and self.flatten_samples:
            if supervised_keys[0] not in self.data_streams:
                # remove <role> of supervised keys
                _, data_stream = split_role_key(supervised_keys[0])
                if not data_stream in self.data_streams:
                    print(
                        "Warning: Cannot find supervised key '{}' in datastreams".format(
                            data_stream
                        )
                    )
                    raise Warning("Unknown data_stream")
                else:
                    supervised_keys[0] = data_stream
            if supervised_keys[1] not in self.schemes:
                # remove <role> of supervised keys
                _, scheme = split_role_key(supervised_keys[1])
                if not scheme in schemes:
                    print(
                        "Warning: Cannot find supervised key '{}' in schemes".format(
                            scheme
                        )
                    )
                    raise Warning("Unknown scheme")
                else:
                    supervised_keys[1] = scheme

        self.supervised_keys = tuple(supervised_keys) if supervised_keys else None

        self._iterable = self._yield_sample()

    def _populate_label_info_from_mongo_doc(self, mongo_schemes):
        """
        Setting self.annos
        Args:
          mongo_schemes:

        Returns:

        """
        annos = {}
        anno_schemes = {}

        # List of all combinations from roles and schemes that occur in the retrieved data.
        for scheme in mongo_schemes:
            for role in self.roles:
                label_id = merge_role_key(role=role, key=scheme["name"])
                scheme_type = nt.string_to_enum(nt.AnnoTypes, scheme["type"])
                scheme_name = scheme["name"]
                scheme_valid = scheme["isValid"]
                anno_schemes[label_id] = scheme_type

                if scheme_type == nt.AnnoTypes.DISCRETE:
                    labels = scheme["labels"]
                    annos[label_id] = nau.DiscreteAnnotation(
                        role=role,
                        add_rest_class=True,
                        scheme=scheme_name,
                        is_valid=scheme_valid,
                        labels=labels,
                    )

                elif scheme_type == nt.AnnoTypes.CONTINUOUS:
                    min_val = scheme["min"]
                    max_val = scheme["max"]
                    sr = scheme["sr"]
                    annos[label_id] = nau.ContinuousAnnotation(
                        role=role,
                        scheme=scheme_name,
                        is_valid=scheme_valid,
                        min_val= min_val,
                        max_val= max_val,
                        sr=sr,
                    )

                elif scheme_type == nt.AnnoTypes.DISCRETE_POLYGON:
                    labels = scheme["labels"]
                    sr = scheme["sr"]
                    annos[label_id] = nau.DiscretePolygonAnnotation(
                        role=role,
                        scheme=scheme_name,
                        is_valid=scheme_valid,
                        labels=labels,
                        sr=sr,
                    )

                elif scheme_type == nt.AnnoTypes.FREE:
                    annos[label_id] = nau.FreeAnnotation(
                        role=role, scheme=scheme_name, is_valid=scheme_valid
                    )

                else:
                    raise ValueError("Invalid label type {}".format(scheme["type"]))

        return annos, anno_schemes

    def _populate_data_info_from_mongo_doc(self, mongo_data_streams):
        """
        Setting self.data
        Args:
          mongo_schemes:

        Returns:

        """
        data_info = {}
        data_schemes = {}

        for data_stream in mongo_data_streams:
            for role in self.roles:
                sample_stream_name = (
                        role + "." + data_stream["name"] + "." + data_stream["fileExt"]
                )
                sample_data_path = os.path.join(
                    self.nova_data_dir,
                    self.dataset,
                    self.sessions[0],
                    sample_stream_name,
                )
                dtype = nt.string_to_enum(nt.DataTypes, data_stream["type"])


                if dtype == nt.DataTypes.VIDEO:
                    data = VideoData(
                        role=role,
                        name=data_stream["name"],
                        file_ext=data_stream["fileExt"],
                        sr=data_stream["sr"],
                        is_valid=data_stream["isValid"],
                        sample_data_path=sample_data_path,
                        lazy_loading=self.lazy_loading,
                    )
                elif dtype == nt.DataTypes.AUDIO:
                    data = AudioData(
                        role=role,
                        name=data_stream["name"],
                        file_ext=data_stream["fileExt"],
                        sr=data_stream["sr"],
                        is_valid=data_stream["isValid"],
                        sample_data_path=sample_data_path,
                        lazy_loading=self.lazy_loading,
                    )
                elif dtype == nt.DataTypes.FEATURE:
                    data = StreamData(
                        role=role,
                        name=data_stream["name"],
                        sr=data_stream["sr"],
                        data_type=dtype,
                        is_valid=data_stream["isValid"],
                        sample_data_path=sample_data_path,
                        lazy_loading=self.lazy_loading,
                    )
                else:
                    raise ValueError("Invalid data type {}".format(data_stream["type"]))

                data_id = merge_role_key(role=role, key=data_stream["name"])
                data_info[data_id] = data
                data_schemes[data_id] = dtype

        return data_info, data_schemes

    def _load_annotation_for_session(self, session, time_to_ms=False):
        for label_id, anno in self.annos.items():
            mongo_anno = self.nova_db_handler.get_annos(
                self.dataset, anno.scheme, session, self.annotator, anno.role
            )
            anno.set_annotation_from_mongo_doc(mongo_anno, time_to_ms=time_to_ms)

    def _open_data_reader_for_session(self, session):
        for data_id, data in self.data_info.items():
            data_path = os.path.join(
                self.nova_data_dir, self.dataset, session, data_id + "." + data.file_ext
            )
            data.open_file_reader(data_path)

    def _yield_sample(self):
        """Yields examples."""

        # Needed to sort the samples later and assure that the order is the same as in nova. Necessary for CML till the tfds can be returned in order.
        sample_counter = 1

        for session in self.sessions:

            # Gather all data we need for this session
            self._load_annotation_for_session(session, time_to_ms=True)
            self._open_data_reader_for_session(session)

            session_info = self.nova_db_handler.get_session_info(self.dataset, session)[
                0
            ]
            dur = session_info["duration"]
            # If are loading any datastreams we check if any datastream is shorter than the duration stored in the database suggests
            if self.data_info:
                dur = min(*[v.dur for k, v in self.data_info.items()], dur)

            if not dur:
                raise ValueError("Session {} has no duration.".format(session))

            dur_ms = int(dur * 1000)

            # If framesize is not specified we return the whole session as one junk
            if self.frame_size_ms is None:
                self.frame_size_ms = min(dur_ms, self.end_ms - self.start_ms)
                self.stride_ms = self.frame_size_ms

            # Starting position of the first frame in seconds
            #c_pos_ms = self.left_context_ms + self.start_ms
            c_pos_ms = max(self.left_context_ms, self.start_ms)

            # Generate samples for this session
            while c_pos_ms + self.stride_ms + self.right_context_ms <= min(
                    self.end_ms, dur_ms
            ):
                frame_start_ms = c_pos_ms - self.left_context_ms
                frame_end_ms = c_pos_ms + self.frame_size_ms + self.right_context_ms


                key = (
                        session
                        + "_"
                        + str(frame_start_ms / 1000)
                        + "_"
                        + str(frame_end_ms / 1000)
                )

                labels_for_frame = [
                    {k: v.get_label_for_frame(frame_start_ms, frame_end_ms)}
                    for k, v in self.annos.items()
                ]
                data_for_frame = []

                for k, v in self.data_info.items():
                    sample = v.get_sample(frame_start_ms, frame_end_ms)
                    if sample.shape[0] == 0:
                        print(f'Sample{frame_start_ms}-{frame_end_ms} is empty')
                        c_pos_ms += self.stride_ms
                        continue
                    data_for_frame.append({k: sample})

                sample_dict = {}

                for l in labels_for_frame:
                    sample_dict.update(l)

                for d in data_for_frame:
                    sample_dict.update(d)

                if self.flatten_samples:

                    # grouping labels and data according to roles
                    for role in self.roles:
                        # filter dictionary to contain values for role
                        sample_dict_for_role = dict(
                            filter(lambda elem: role in elem[0], sample_dict.items())
                        )

                        # remove role from dictionary keys
                        sample_dict_for_role = dict(
                            map(
                                lambda elem: (elem[0].replace(role + ".", ""), elem[1]),
                                sample_dict_for_role.items(),
                            )
                        )

                        sample_dict_for_role["frame"] = (
                                str(sample_counter) + "_" + key + "_" + role
                        )
                        # yield key + '_' + role, sample_dict_for_role
                        yield sample_dict_for_role
                        sample_counter += 1
                    c_pos_ms += self.stride_ms

                else:
                    sample_dict["frame"] = str(sample_counter) + "_" + key
                    yield sample_dict
                    c_pos_ms += self.stride_ms
                    sample_counter += 1

            # closing file readers for this session
            for k, v in self.data_info.items():
                v.close_file_reader()

    def __iter__(self):
        return self._iterable

    def __next__(self):
        return self._iterable.__next__()

    def get_output_info(self):

        def map_label_id(lid):
            if self.flatten_samples and not lid == 'frame':
                return split_role_key(lid)[-1]
            return lid

        return {
                    # Adding fake framenumber label for sorting
                    'frame': {"dtype":np.str, "shape":(1,)},
                    **{map_label_id(k): v.get_info()[1] for k, v in self.annos.items()},
                    **{map_label_id(k): v.get_info()[1] for k, v in self.data_info.items()}
                }




import cv2
from decord import VideoReader, AudioReader, cpu
import numpy as np
import os

# import librosa
from hcai_datasets.hcai_nova_dynamic.utils import nova_types as nt
from hcai_datasets.hcai_nova_dynamic.utils.ssi_stream_utils import Stream
from hcai_datasets.hcai_nova_dynamic.utils.nova_utils import (
    merge_role_key,
    split_role_key,
)
from typing import Union
from abc import ABC, abstractmethod


class Data(ABC):

    def __init__(
            self,
            role: str = "",
            name: str = "",
            file_ext: str = "stream",
            sr: int = 0,
            data_type: nt.DataTypes = None,
            is_valid: bool = True,
            sample_data_path: str = "",
            lazy_loading: bool = False,
    ):
        self.role = role
        self.name = name
        self.is_valid = is_valid
        self.sr = sr
        self.file_ext = file_ext
        self.lazy_loading = lazy_loading
        self.data_type = data_type

        # Set when populate_meta_info is called
        self.sample_data_shape = None
        self.np_data_type = None
        self.meta_loaded = False

        # Set when open_file_reader is called
        self.file_path = None
        self.file_reader = None
        self.dur = None

        if sample_data_path:
            self.populate_meta_info(sample_data_path)

    def data_stream_opend(self):
        if not self.file_reader:
            print(
                "No datastream opened for {}".format(
                    merge_role_key(self.role, self.name)
                )
            )
            raise RuntimeError("Datastream not loaded")

    def get_info(self):
        if self.meta_loaded:
            if self.lazy_loading:
                return {
                    "frame_start": {"dtype": np.float32, "shape": (1)},
                    "frame_end": {"dtype": np.float32, "shape": (1)},
                    "file_path": {"dtype": np.str, "shape": (None,)}
                }
            return self.get_info_hook()
        else:
            print(
                "Meta data has not been loaded for file {}. Call get_meta_info() first.".format(
                    merge_role_key(self.role, self.name)
                )
            )

    def get_sample(self, frame_start: int, frame_end: int):
        """
        Returns the sample for the respective frames. If lazy loading is true, only the filepath and frame_start, frame_end will be returned.
        """
        if self.lazy_loading:
            return {
                "frame_start": frame_start,
                "frame_end": frame_end,
                "file_path": self.file_path,
            }
        else:
            return self.get_sample_hook(frame_start, frame_end)

    def open_file_reader(self, path: str):
        self.file_path = path
        self.open_file_reader_hook(path)

    @abstractmethod
    def get_info_hook(self):
        """
        Returns the features for this datastream to create the DatasetInfo for tensorflow
        """
        ...

    @abstractmethod
    def get_sample_hook(self, start_frame: int, end_frame: int):
        """
        Returns a data chunk from start frame to end frame
        """
        ...

    @abstractmethod
    def open_file_reader_hook(self, path: str):
        """
        Opens a filereader for the respective datastream. Sets attributes self.file_reader and self.dur
        """
        ...

    @abstractmethod
    def populate_meta_info(self, path: str):
        """
        Opens a data sample from the provided path to extract additional data that is not in the database
        """
        ...

    @abstractmethod
    def close_file_reader(self):
        """
        Closes a filereader for the respective datastream
        """
        ...


class AudioData(Data):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get_info_hook(self):
        return merge_role_key(self.role, self.name), {
            "shape": self.sample_data_shape,
            "dtype": np.float32
        }

    def get_sample_hook(self, frame_start_ms: int, frame_end_ms: int):
        start_frame = int(frame_start_ms * self.sr / 1000)
        end_frame = int(frame_end_ms * self.sr / 1000)
        chunk = self.file_reader.get_batch(
            list(range(start_frame, end_frame))
        ).asnumpy()
        return chunk

    def open_file_reader_hook(self, path: str):
        if not os.path.isfile(path):
            print("Session file not found at {}.".format(path))
            raise FileNotFoundError()
        self.file_reader = AudioReader(path, ctx=cpu(0), mono=False)
        self.dur = self.file_reader.duration()

    def populate_meta_info(self, path: str):
        if not os.path.isfile(path):
            print("Sample file not found at {}. Can't load metadata.".format(path))
            raise FileNotFoundError()
        file_reader = AudioReader(path, ctx=cpu(0), mono=False)
        n_channels = file_reader.shape[0]
        self.sample_data_shape = (None, n_channels)
        self.meta_loaded = True

    def close_file_reader(self):
        return True


class VideoData(Data):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get_info_hook(self):
        return merge_role_key(self.role, self.name), {
            "shape": self.sample_data_shape,
            "dtype": np.float32
        }

    def get_sample_hook(self, frame_start_ms: int, frame_end_ms: int):
        start_frame = int(frame_start_ms * self.sr / 1000)
        end_frame = int(frame_end_ms * self.sr / 1000)
        chunk = self.file_reader.get_batch(
            list(range(start_frame, end_frame))
        ).asnumpy()
        return chunk

    def open_file_reader_hook(self, path: str):
        if not os.path.isfile(path):
            print("Session file not found at {}.".format(path))
            raise FileNotFoundError()
        self.file_reader = VideoReader(path, ctx=cpu(0))
        fps = self.file_reader.get_avg_fps()
        frame_count = len(self.file_reader)
        self.dur = frame_count / fps

    def populate_meta_info(self, path: str):
        if not os.path.isfile(path):
            print("Sample file not found at {}. Can't load metadata.".format(path))
            raise FileNotFoundError()
        file_reader = VideoReader(path)
        self.sample_data_shape = file_reader[0].shape
        self.meta_loaded = True

    def close_file_reader(self):
        return True


class StreamData(Data):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get_info_hook(self):
        return merge_role_key(self.role, self.name), {
            "shape": self.sample_data_shape,
            "dtype": self.np_data_type
        }

    def get_sample_hook(self, frame_start_ms: int, frame_end_ms: int):
        try:
            self.data_stream_opend()
            start_frame = milli_seconds_to_frame(self.sr, frame_start_ms)
            end_frame = milli_seconds_to_frame(self.sr, frame_end_ms)
            return self.file_reader.data[start_frame:end_frame]
        except RuntimeError:
            print(
                "Could not get chunk {}-{} from data stream {}".format(
                    frame_start_ms, frame_end_ms, merge_role_key(self.role, self.name)
                )
            )

    def open_file_reader_hook(self, path: str) -> bool:
        stream = Stream(path)
        if stream:
            self.file_reader = stream
            self.dur = stream.data.shape[0] / stream.sr
            return True
        else:
            print("Could not open Stream {}".format(str))
            return False

    def close_file_reader(self):
        return True

    def populate_meta_info(self, path: str):
        stream = Stream().load_header(path)
        self.sample_data_shape = (stream.dim,)
        self.np_data_type = stream.type
        self.meta_loaded = True


##########################
# General helper functions
##########################


def frame_to_seconds(sr: int, frame: int) -> float:
    return frame / sr


def seconds_to_frame(sr: int, time_s: float) -> int:
    return round(time_s * sr)


def milli_seconds_to_frame(sr: int, time_ms: int) -> int:
    return seconds_to_frame(sr=sr, time_s=time_ms / 1000)


def parse_time_string_to_ms(frame: Union[str, int, float]) -> int:
    # if frame is specified milliseconds as string
    if str(frame).endswith("ms"):
        try:
            return float(frame[:-2])
        except ValueError:
            raise ValueError(
                "Invalid input format for frame in milliseconds: {}".format(frame)
            )
    # if frame is specified in seconds as string
    elif str(frame).endswith("s"):
        try:
            frame_s = float(frame[:-1])
            return float(frame_s * 1000)
        except ValueError:
            raise ValueError(
                "Invalid input format for frame in seconds: {}".format(frame)
            )
    # if type is float we assume the input will be seconds
    elif isinstance(frame, float) or "." in str(frame):
        try:
            print(
                "WARNING: Automatically inferred type for frame {} is float.".format(
                    frame
                )
            )
            return float(1000 * float(frame))
        except ValueError:
            raise ValueError("Invalid input format for frame: {}".format(frame))
    # if type is int we assume the input will be milliseconds
    elif isinstance(frame, int) or (isinstance(frame, str) and frame.isdigit()):
        try:
            print(
                "WARNING: Automatically inferred type for frame {} is int.".format(
                    frame
                )
            )
            return float(frame)
        except ValueError:
            raise ValueError("Invalid input format for frame: {}".format(frame))

    print(f'WARNING: Could  not automatically parse time "{frame}" to seconds. Returning None ')


import os
import copy
import warnings
import configparser

from typing import Union
from datetime import datetime
from pymongo import MongoClient
from pymongo.results import InsertOneResult, UpdateResult


class NovaDBHandler:
    ANNOTATOR_COLLECTION = "Annotators"
    SCHEME_COLLECTION = "Schemes"
    STREAM_COLLECTION = "Streams"
    ROLE_COLLECTION = "Roles"
    ANNOTATION_COLLECTION = "Annotations"
    SESSION_COLLECTION = "Sessions"
    ANNOTATION_DATA_COLLECTION = "AnnotationData"

    def __init__(self, db_config_path=None, db_config_dict=None):

        # Connecting to the database
        if db_config_path:
            if os.path.isfile(db_config_path):
                cfg = self.read_config(db_config_path)
                self.ip = str(cfg["DB"]["ip"])
                self.port = int(cfg["DB"]["port"])
                self.user = str(cfg["DB"]["user"])
                self.password = str(cfg["DB"]["password"])
            else:
                raise FileNotFoundError(
                    "No database config file found at {}".format(db_config_path)
                )

        # If a db config_dict is specified overwrite the config from path
        if db_config_dict:
            if db_config_path:
                print(
                    "WARNING! database config are specifed as file AND and as dictionary. Using the dictionary."
                )
            self.ip = db_config_dict["ip"]
            self.port = db_config_dict["port"]
            self.user = db_config_dict["user"]
            self.password = db_config_dict["password"]

        if not (self.ip or self.port or self.user or self.password):
            print(
                "WARNING! No valid nova database config found for path {} and dict {} \n Found config parameters are ip:{}, port{}, user: {}. Also check your password.".format(
                    db_config_path, db_config_dict, self.ip, self.port, self.user
                )
            )

        self.client = MongoClient(
            host=self.ip, port=self.port, username=self.user, password=self.password
        )

    def print_config(self, cfg, cfg_path):
        print("Loaded config from {}:".format(cfg_path))
        print("---------------------")
        for sec_name, sec_dict in cfg._sections.items():
            print(sec_name)
            for k, v in sec_dict.items():
                if k == "password":
                    continue
                else:
                    print("\t{} : {}".format(k, v))
        print("---------------------")

    def read_config(self, cfg_path):
        config = configparser.RawConfigParser()
        config.read(cfg_path)
        self.print_config(config, cfg_path)
        return config

    # Reading from database
    def get_docs_by_prop(
            self, vals: Union[list, str], property: str, database: str, collection: str
    ) -> list:
        """
        Fetching a document from the mongo db collection in the respective database where the passed values are matching a specific property in the collection.

        Args:
          vals: The specific value of a property in the document
          property: The property to look for the passed values
          database: The name of the database to search
          collection: The name of the collection within the database to search

        Returns:
          list: List of documents that match the specified criteria

        """
        filter = []

        if not type(vals) == type(list()):
            vals = [vals]

        for n in vals:
            filter.append({property: n})

        filter = {"$or": filter}
        ret = list(self.client[database][collection].find(filter))
        return ret

    def get_schemes(self, dataset, schemes):
        """
        Fetches the scheme object that matches the specified criteria from the nova database and returns them as a python readable dictionary.
        Args:
          ip:
          port:
          user:
          password:
          dataset:
          scheme:

        Returns:

        """

        if not schemes:
            print("WARNING: No Schemes have been requested. Returning empty list.")
            return []

        # if not dataset in self.datasets:
        # raise ValueError('{} not found in datasets'.format(dataset))

        mongo_schemes = []
        for scheme in schemes:
            mongo_scheme = self.get_docs_by_prop(
                scheme, "name", dataset, self.SCHEME_COLLECTION
            )
            if not mongo_scheme:
                print(f"WARNING: No scheme {scheme} found in database for dataset {dataset}")
            else:
                mongo_schemes.append(mongo_scheme[0])

        if not mongo_schemes:
            raise ValueError(
                "No entries for schemes {} found in database".format(schemes)
            )

        return mongo_schemes

    def get_session_info(self, dataset, session):
        """
        Fetches the session object that matches the specified criteria from the nova database and returns them as a python readable dictionary.

        Args:
          dataset:
          session:

        Returns:

        """
        mongo_session = self.get_docs_by_prop(
            session, "name", dataset, self.SESSION_COLLECTION
        )
        return mongo_session

    def get_data_streams(self, dataset, data_streams):
        """
        Fetches the stream objects that matches the specified criteria from the nova database and returns them as a python readable dictionary.
        Args:
          dataset:
          session:
          role_list:
          data_stream_list:
        """
        # if not dataset in self.datasets:
        #  raise ValueError('{} not found in datasets'.format(dataset))

        if not data_streams:
            print("WARNING: No Datastreams have been requested. Returning empty list.")
            return []

        mongo_streams = []
        for stream in data_streams:
            mongo_stream = self.get_docs_by_prop(
                stream, "name", dataset, self.STREAM_COLLECTION
            )
            if not mongo_stream:
                print("WARNING: No stream {} found in database".format(stream))
            else:
                mongo_streams.append(mongo_stream[0])

        if not mongo_streams:
            raise ValueError("no entries for datastream {} found".format(data_streams))

        return mongo_streams

    def get_annotation_docs(
            self,
            mongo_schemes,
            mongo_sessions,
            mongo_annotators,
            mongo_roles,
            database,
            collection,
    ):
        """
        Fetches all annotationobjects that match the specified criteria from the nova database
        Args:
          mongo_schemes:
          mongo_sessions:
          mongo_annotators:
          mongo_roles:
          database:
          collection:
          client:

        Returns:

        """
        scheme_filter = []
        role_filter = []
        annotator_filter = []
        session_filter = []

        for ms in mongo_schemes:
            scheme_filter.append({"scheme_id": ms["_id"]})

        for mse in mongo_sessions:
            session_filter.append({"session_id": mse["_id"]})

        for ma in mongo_annotators:
            annotator_filter.append({"annotator_id": ma["_id"]})

        for mr in mongo_roles:
            role_filter.append({"role_id": mr["_id"]})

        filter = {
            "$and": [
                {"$or": scheme_filter},
                {"$or": session_filter},
                {"$or": role_filter},
                {"$or": annotator_filter},
            ]
        }

        ret = list(self.client[database][collection].find(filter))
        return ret

    def get_annos(
            self,
            dataset: str,
            scheme: str,
            session: str,
            annotator: str,
            roles: Union[list, str],
    ) -> list:
        """
        Fetches all annotations that matches the specified criteria from the nova database and returns them as a list of python readable dictionaries.
        Args:
          dataset:
          scheme:
          session:
          annotator:
          roles:

        Returns:

        """
        mongo_schemes = self.get_docs_by_prop(
            scheme, "name", dataset, self.SCHEME_COLLECTION
        )
        if not mongo_schemes:
            warnings.warn(f"Unknown scheme {scheme} found")
            return []
        mongo_annotators = self.get_docs_by_prop(
            annotator, "name", dataset, self.ANNOTATOR_COLLECTION
        )
        if not mongo_annotators:
            warnings.warn(f"Unknown annotator {annotator} found")
            return []
        mongo_roles = self.get_docs_by_prop(roles, "name", dataset, self.ROLE_COLLECTION)
        if not mongo_roles:
            warnings.warn(f"Unknown role {roles} found")
            return []
        mongo_sessions = self.get_docs_by_prop(
            session, "name", dataset, self.SESSION_COLLECTION
        )
        if not mongo_sessions:
            warnings.warn(f"Unknown for session {session} found")
            return []

        mongo_annos = self.get_annotation_docs(
            mongo_schemes,
            mongo_sessions,
            mongo_annotators,
            mongo_roles,
            dataset,
            self.ANNOTATION_COLLECTION,
        )

        # getting the annotation data and the session name
        if not mongo_annos:
            print(
                f"No annotations found for \n\t-annotator: {annotator}\n\t-scheme: {scheme}\n\t-session: {session}\n\t-role: {roles}"
            )
            return []

        else:
            # TODO: adapt for multiple roles, annotators etc.
            label = self.get_data_docs_by_prop(
                mongo_annos[0]["data_id"], "_id", dataset
            )
            label = label["labels"]

        return label

    def insert_doc_by_prop(
            self, doc: dict, database: str, collection: str
    ) -> InsertOneResult:
        """
        Uploading a document to the database using the specificed parameters
        Args:
          docs: List of dictionaries with objects to upload to the database
          database: The name of the database to search
          collection: The name of the collection within the database to search

        Returns:
          str: ObjectID of the inserted objects or an empty list in case of failure
        """
        ret = self.client[database][collection].insert_one(doc)
        return ret

    def update_doc_by_prop(
            self, doc: dict, database: str, collection: str
    ) -> UpdateResult:
        """
        Uploading a document to the database using the specificed parameters
        Args:
          docs: List of dictionaries with objects to upload to the database
          database: The name of the database to search
          collection: The name of the collection within the database to search

        Returns:
          str: ObjectID of the inserted objects or an empty list in case of failure
        """
        ret = self.client[database][collection].update_one(
            {"_id": doc["_id"]}, {"$set": doc}
        )
        return ret

    def update_doc_by_id(
            self, _id: str, doc: dict, database: str, collection: str
    ) -> UpdateResult:
        """
        Uploading a document to the database using the specificed parameters
        Args:
          _id: ID of doc, which has to be updated
          docs: List of dictionaries with objects to upload to the database
          database: The name of the database to search
          collection: The name of the collection within the database to search

        Returns:
          str: ObjectID of the inserted objects or an empty list in case of failure
        """
        ret = self.client[database][collection].update_one(
            {"_id": _id}, {"$set": doc}
        )
        return ret

    # TODO: Remove Restclass Labels in discrete Cases
    # TODO: Consider "forced overwrite"
    # TODO: Add Backup case
    # TODO: Call preprocess of annotation
    def set_annos(
            self,
            database: str,
            scheme: str,
            session: str,
            annotator: str,
            role: str,
            annos: list,
            is_finished: bool = False,
            is_locked: bool = False,
    ) -> str:
        """
        Uploading annotations to the database
        Args:
          database:
          scheme:
          session:
          annotator:
          role:
          annos:

        Returns: Object ID of the inserted annotations. Empty string in case of failure
        """
        mongo_scheme = self.get_mongo_scheme(scheme, database)
        mongo_annotator = self.get_mongo_annotator(annotator, database)
        mongo_role = self.get_mongo_role(role, database)
        mongo_session = self.get_mongo_session(session, database)

        # Check if annotations already exist
        mongo_annos = self.get_annotation_docs(
            mongo_scheme,
            mongo_session,
            mongo_annotator,
            mongo_role,
            database,
            self.ANNOTATION_COLLECTION,
        )

        # Check for existing annotations
        mongo_anno_id = None
        mongo_data_id = None
        if mongo_annos:
            if mongo_annos[0]["isLocked"]:
                warnings.warn(f"Can't overwrite locked annotation {str(mongo_annos[0]['_id'])}")
                return ""
            else:
                warnings.warn(f"Overwriting existing annotation {str(mongo_annos[0]['_id'])}")
                mongo_anno_id = mongo_annos[0]["_id"]
                mongo_data_id = mongo_annos[0]["data_id"]

        # Upload label data
        mongo_label_doc = {"labels": annos}
        if mongo_data_id:
            mongo_label_doc["_id"] = mongo_data_id
            success = self.update_doc_by_prop(
                doc=mongo_label_doc,
                database=database,
                collection=self.ANNOTATION_DATA_COLLECTION,
            )
            if not success.acknowledged:
                warnings.warn(f"Unknown error update database entries for Annotation data {mongo_data_id}")
                return ""
            else:
                data_id = mongo_data_id
        else:
            success = self.insert_doc_by_prop(
                doc=mongo_label_doc,
                database=database,
                collection=self.ANNOTATION_DATA_COLLECTION,
            )
            if not success.acknowledged:
                warnings.warn(f"Unexpected error uploading annotation data for {database} - {session} - {scheme} - "
                              f"{annotator}. Upload failed.")
                return ""
            else:
                data_id = success.inserted_id

        # Upload annotation information
        mongo_anno_doc = {
            "data_id": data_id,
            "annotator_id": mongo_annotator[0]["_id"],
            "role_id": mongo_role[0]["_id"],
            "scheme_id": mongo_scheme[0]["_id"],
            "session_id": mongo_session[0]["_id"],
            "isFinished": is_finished,
            "isLocked": is_locked,
            "date": datetime.today().replace(microsecond=0),
        }

        if mongo_anno_id:
            mongo_anno_doc["_id"] = mongo_anno_id
            success = self.update_doc_by_prop(
                doc=mongo_anno_doc, database=database, collection=self.ANNOTATION_COLLECTION
            )
            if not success.acknowledged:
                warnings.warn(
                    f"Unexpected error uploading annotations for {database} - {session} - {scheme} - {annotator}. Upload failed."
                )
                return ""
            else:
                anno_id = mongo_anno_id
        else:
            success = self.insert_doc_by_prop(
                doc=mongo_anno_doc, database=database, collection=self.ANNOTATION_COLLECTION
            )
            if not success.acknowledged:
                warnings.warn(
                    f"Unexpected error uploading annotations for {database} - {session} - {scheme} - {annotator}. Upload failed."
                )
                return ""
            else:
                anno_id = success.inserted_id
        return anno_id

    def get_mongo_scheme(self, scheme, database):
        mongo_scheme = self.get_docs_by_prop(scheme, "name", database, self.SCHEME_COLLECTION)
        if not mongo_scheme:
            warnings.warn(f"Unknown scheme {scheme} found")
            return ""

        return mongo_scheme

    def get_mongo_annotator(self, annotator, database):
        mongo_annotator = self.get_docs_by_prop(annotator, "name", database, self.ANNOTATOR_COLLECTION)
        if not mongo_annotator:
            warnings.warn(f"Unknown annotator {annotator} found")
            return ""

        return mongo_annotator

    def get_mongo_role(self, role, database):
        mongo_role = self.get_docs_by_prop(role, "name", database, self.ROLE_COLLECTION)
        if not mongo_role:
            warnings.warn(f"Unknown role {role} found")
            return ""

        return mongo_role

    def get_mongo_session(self, session, database):
        mongo_session = self.get_docs_by_prop(session, "name", database, self.SESSION_COLLECTION)
        if not mongo_session:
            warnings.warn(f"Unknown for session {session} found")
            return ""

        return mongo_session

    def delete_doc_with_tail(self, doc_id_to_remove, database):
        while doc_id_to_remove is not None:
            remove_id = copy.deepcopy(doc_id_to_remove)
            result = self.get_fields_by_properties(doc_id_to_remove, "_id", "nextEntry", database,
                                                   self.ANNOTATION_DATA_COLLECTION)
            if result is not None and 'nextEntry' in result:
                doc_id_to_remove = result['nextEntry']
            else:
                doc_id_to_remove = None

            self.delete_doc_by_prop(remove_id, "_id", database, self.ANNOTATION_DATA_COLLECTION)

    def delete_doc_by_prop(self, vals: Union[list, str], property: str, database: str, collection):
        filter = []

        if not isinstance(vals, list):
            vals = [vals]

        for n in vals:
            filter.append({property: n})

        filter = {"$and": filter}

        return self.client[database][collection].delete_one(filter)

    def get_data_docs_by_prop(self, vals: Union[list, str], property: str, database: str):
        filter = []

        if not isinstance(vals, list):
            vals = [vals]

        for n in vals:
            filter.append({property: n})

        filter = {"$or": filter}

        result = list(self.client[database][self.ANNOTATION_DATA_COLLECTION].find(filter))[0]
        if "nextEntry" in result:
            return self.merge_collections(result, database)

        return result

    def get_fields_by_properties(self, vals: Union[list, str], property, fields: Union[list, str], database: str,
                                 collection: str):
        filter = []
        fields_dict = {}

        if not isinstance(fields, list):
            fields = [fields]

        for n in fields:
            fields_dict[n] = 1

        if not isinstance(vals, list):
            vals = [vals]

        for n in vals:
            filter.append({property: n})

        filter = {"$or": filter}

        return self.client[database][collection].find_one(filter, fields_dict)

    def merge_collections(self, doc, database):
        next_id = doc["nextEntry"]
        new_doc = self.get_data_docs_by_prop(next_id, "_id", database)
        doc['labels'] += new_doc['labels']

        return doc


if __name__ == "__main__":
    db_handler = NovaDBHandler("../../local/nova_db_test.cfg")

    test_cont = False
    test_cat = False
    test_free = True

    # Test continuous data download and upload
    if test_cont:
        dataset = "aria-noxi"
        session = "004_2016-03-18_Paris"
        scheme = "engagement"
        annotator = "system"
        roles = ["novice"]

        mongo_scheme = db_handler.get_schemes(dataset=dataset, schemes=[scheme])
        annos = db_handler.get_annos(
            dataset=dataset,
            scheme=scheme,
            session=session,
            annotator=annotator,
            roles=roles,
        )

    # Test categorical data download and upload
    if test_cat:
        dataset = "roxi"
        session = "001"
        scheme = "emotionalbursts"
        annotator = "gold"
        roles = ["player1"]

        mongo_scheme = db_handler.get_schemes(dataset=dataset, schemes=[scheme])
        annos = db_handler.get_annos(
            dataset=dataset,
            scheme=scheme,
            session=session,
            annotator=annotator,
            roles=roles,
        )

        new_annotator = "test"
        new_annos = [
            {"from": 0, "to": 10, "id": 1, "conf": 0.5},
            {"from": 20, "to": 25, "id": 1, "conf": 1},
            {"from": 30, "to": 35, "id": 1, "conf": 1},
        ]

        db_handler.set_annos(
            dataset=dataset,
            scheme=scheme,
            session=session,
            annotator=new_annotator,
            role=roles[0],
            annos=new_annos,
        )

    # Test free label download and upload
    if test_free:
        dataset = "kassel_therapie_korpus"
        session = "OPD_102"
        scheme = "transcript"
        annotator = "system"
        roles = ["therapist"]

        mongo_scheme = db_handler.get_schemes(dataset=dataset, schemes=[scheme])
        annos = db_handler.get_annos(
            dataset=dataset,
            scheme=scheme,
            session=session,
            annotator=annotator,
            roles=roles,
        )

        new_annotator = "schildom"
        new_annos = [
            {"from": 0, "to": 10, "conf": 1, "name": 'das'},
            {"from": 20, "to": 25, "conf": 1, "name": 'geht'},
            {"from": 30, "to": 35, "conf": 1, "name": 'ja'},
        ]

        db_handler.set_annos(
            dataset=dataset,
            scheme=scheme,
            session=session,
            annotator=new_annotator,
            role=roles[0],
            annos=new_annos,
        )

    print("Done")
